import torch
import torch.nn as nn
import torch.nn.functional as F

class Lateral(nn.Module):

    def __init__(self, in_channels, out_channels, num_layers, bn=False, dilation=1):
        super(Lateral, self).__init__()

        self.num_layers = num_layers
        self.conv = nn.ModuleList()
        for i in range(num_layers):
            conv = []
            if i == num_layers - 2:
                conv.append(nn.Conv2d(in_channels, out_channels, kernel_size=3, stride=1, padding=dilation, dilation=dilation))
                in_channels = out_channels
            else:
                conv.append(nn.Conv2d(in_channels, in_channels, kernel_size=3, stride=1, padding=dilation, dilation=dilation))
            if bn:
                conv.append(nn.BatchNorm2d(in_channels))
            conv.append(nn.ReLU(inplace=True))
            self.conv.append(nn.Sequential(*conv))

        self.initialize()

    def initialize(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_uniform_(m.weight.data, a=1)
                if m.bias is not None:
                    m.bias.data.zero_()
    def forward(self, x):
        for i in range(self.num_layers):        
            x = self.conv[i](x)
        return x


class CLIFF(nn.Module):
    def __init__(self, in_channels, out_channels, bn=False):
        super(CLIFF, self).__init__()
        if bn:
            self.conv1 = nn.Sequential(nn.Conv2d(2*in_channels, out_channels, 3, padding=1, bias=False),
                                      nn.BatchNorm2d(out_channels),
                                      nn.ReLU(inplace=True))
            self.conv2 = nn.Sequential(nn.Conv2d(2*in_channels, out_channels, 3, padding=1, bias=False),
                                      nn.BatchNorm2d(out_channels),
                                      nn.ReLU(inplace=True))
        else:
            self.conv1 = nn.Sequential(nn.Conv2d(2*in_channels, out_channels, 3, padding=1),
                                      nn.ReLU(inplace=True))
            self.conv2 = nn.Sequential(nn.Conv2d(2*in_channels, out_channels, 3, padding=1),
                                      nn.ReLU(inplace=True))
        self.initialize()

    def initialize(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_uniform_(m.weight.data, a=1)
                if m.bias is not None:
                    m.bias.data.zero_()
        
    def forward(self, x_l, x_h):
        x_h = F.interpolate(x_h, size=x_l.shape[2:], mode='bilinear', align_corners=False)
        x_la = x_l * x_h
        x_lc = torch.cat([x_l, x_la], dim=1)
        x_lc = self.conv1(x_lc)
        x_c = torch.cat([x_h, x_lc], dim=1)
        y = self.conv2(x_c)
        out = y + x_h
        return out


class DPN(nn.Module):

    def __init__(self, feat_channels=256, num_layers=1, bn=True):
        super(DPN, self).__init__()
        self.lateral2 = Lateral(feat_channels, feat_channels, num_layers, bn=bn)
        self.lateral3 = Lateral(feat_channels, feat_channels, num_layers, bn=bn)
        self.lateral4 = Lateral(feat_channels, feat_channels, num_layers, bn=bn)
        self.lateral5 = Lateral(feat_channels, feat_channels, num_layers, bn=bn)
        self.cliff4 = CLIFF(feat_channels, feat_channels, bn=bn)
        self.cliff3 = CLIFF(feat_channels, feat_channels, bn=bn)
        self.cliff2 = CLIFF(feat_channels, feat_channels, bn=bn)

        if bn:
            self.score = nn.Sequential(nn.Conv2d(feat_channels, feat_channels, 3, padding=1, bias=False), 
                                       nn.BatchNorm2d(feat_channels), 
                                       nn.ReLU(inplace=True),
                                       nn.Conv2d(feat_channels, 1, 3, padding=1),
                                  )
        else:
            self.score = nn.Sequential(nn.Conv2d(feat_channels, feat_channels, 3, padding=1), 
                                       nn.ReLU(inplace=True),
                                       nn.Conv2d(feat_channels, 1, 3, padding=1),
                                  )

        #self.initialize()

    def forward(self, fpn_p2, fpn_p3, fpn_p4, fpn_p5, img):
        fpn_p2 = self.lateral2(fpn_p2)
        fpn_p3 = self.lateral3(fpn_p3)
        fpn_p4 = self.lateral4(fpn_p4)
        fpn_p5 = self.lateral5(fpn_p5)

        f4 = self.cliff4(fpn_p4, fpn_p5)
        f3 = self.cliff3(fpn_p3, f4)
        f2 = self.cliff2(fpn_p2, f3)
        score_lr = self.score(f2)
        score = F.interpolate(score_lr, img.shape[2:], mode='bilinear', align_corners=False)
        return score


    def initialize(self):
        nn.init.normal_(self.score.weight.data, 0, 0.01)
        self.score.bias.data.zero_()


